@@ -1,12 +1,13 @@
+from functools import partial
 import torch
 from torch import nn, einsum
+from torch.autograd import grad
 import torch.nn.functional as F
 
 from einops import rearrange, reduce, repeat
 from einops.layers.torch import Rearrange, Reduce
 
 from vector_quantize_pytorch import VectorQuantize as VQ
-from axial_positional_embedding import AxialPositionalEmbedding
 
 import torchvision
 
@@ -24,6 +25,31 @@ def exists(val):
 def default(val, d):
     return val if exists(val) else d
 
+# keyword argument helpers
+
+def pick_and_pop(keys, d):
+    values = list(map(lambda key: d.pop(key), keys))
+    return dict(zip(keys, values))
+
+def group_dict_by_key(cond, d):
+    return_val = [dict(),dict()]
+    for key in d.keys():
+        match = bool(cond(key))
+        ind = int(not match)
+        return_val[ind][key] = d[key]
+    return (*return_val,)
+
+def string_begins_with(prefix, str):
+    return str.startswith(prefix)
+
+def group_by_key_prefix(prefix, d):
+    return group_dict_by_key(partial(string_begins_with, prefix), d)
+
+def groupby_prefix_and_trim(prefix, d):
+    kwargs_with_prefix, kwargs = group_dict_by_key(partial(string_begins_with, prefix), d)
+    kwargs_without_prefix = dict(map(lambda x: (x[0][len(prefix):], x[1]), tuple(kwargs_with_prefix.items())))
+    return kwargs_without_prefix, kwargs
+
 # decorators
 
 def eval_decorator(fn):
@@ -50,6 +76,9 @@ def gumbel_noise(t):
 def gumbel_sample(t, temperature = 1., dim = -1):
     return ((t / temperature) + gumbel_noise(t)).argmax(dim = dim)
 
+def safe_div(numer, denom, eps = 1e-6):
+    return numer / (denom + eps)
+
 # gan losses
 
 def hinge_discr_loss(fake, real):
@@ -64,6 +93,14 @@ def bce_discr_loss(fake, real):
 def bce_gen_loss(fake):
     return -log(sigmoid(fake)).mean()
 
+def grad_layer_wrt_loss(loss, layer):
+    return grad(
+        outputs = loss,
+        inputs = layer,
+        grad_outputs = torch.ones_like(loss),
+        retain_graph = True
+    )[0].detach()
+
 # vqgan vae
 
 class Discriminator(nn.Module):
@@ -105,16 +142,19 @@ class VQGanVAE(nn.Module):
         vq_decay = 0.8,
         vq_commitment_weight = 1.,
         l2_recon_loss = False,
-        use_hinge_loss = False
+        use_hinge_loss = False,
+        **kwargs
     ):
         super().__init__()
+        vq_kwargs, kwargs = groupby_prefix_and_trim('vq_', kwargs)
+
         self.num_layers = num_layers
         self.codebook_size = vq_codebook_size
 
         self.encoders = MList([])
         self.decoders = MList([])
 
-        dims = (channels, *((dim,) * num_layers))
+        dims = (dim,) * num_layers
         reversed_dims = tuple(reversed(dims))
         enc_dim_pairs = zip(dims[:-1], dims[1:])
         dec_dim_pairs = zip(reversed_dims[:-1], reversed_dims[1:])
@@ -123,17 +163,21 @@ class VQGanVAE(nn.Module):
             self.encoders.append(nn.Conv2d(enc_dim_in, enc_dim_out, 4, stride = 2, padding = 1))
             self.decoders.append(nn.ConvTranspose2d(dec_dim_in, dec_dim_out, 4, stride = 2, padding = 1))
 
+        self.encoders.insert(0, nn.Conv2d(channels, dim, 3, padding = 1))
+        self.decoders.append(nn.Conv2d(dim, channels, 1))
+
         self.vq = VQ(
             dim = dim,
             codebook_size = vq_codebook_size,
             decay = vq_decay,
             commitment_weight = vq_commitment_weight,
-            accept_image_fmap = True
+            accept_image_fmap = True,
+            **vq_kwargs
         )
 
         # reconstruction loss
 
-        self.l2_recon_loss = l2_recon_loss
+        self.recon_loss_fn = F.mse_loss if l2_recon_loss else F.l1_loss
 
         # preceptual loss
 
@@ -188,29 +232,39 @@ class VQGanVAE(nn.Module):
 
         assert return_loss ^ return_discr_loss, 'you should either return autoencoder loss or discriminator loss, but not both'
 
-        if return_discr_loss:
-            fmap.detach_()
-            fmap_discr_logits, img_discr_logits = map(self.discr, (fmap, img))
-            return self.discr_loss(fmap_discr_logits, img_discr_logits)
-
-        # generator loss
-
-        gen_loss = self.gen_loss(fmap)
-
-        # reconstruction loss
-
-        recon_loss_fn = F.mse_loss if self.l2_recon_loss else F.l1_loss
-        recon_loss = recon_loss_fn(fmap, img)
-
         # lpips
 
         img_vgg_feats = self.vgg(img)
         recon_vgg_feats = self.vgg(fmap)
         perceptual_loss = F.mse_loss(img_vgg_feats, recon_vgg_feats)
 
+        # whether to return discriminator loss
+
+        if return_discr_loss:
+            fmap.detach_()
+            fmap_discr_logits, img_discr_logits = map(self.discr, (fmap, img))
+            discr_loss = self.discr_loss(fmap_discr_logits, img_discr_logits)
+            return discr_loss
+
+        # generator loss
+
+        gen_loss = self.gen_loss(fmap)
+
+        # calculate adaptive weight
+
+        last_dec_layer = self.decoders[-1].weight
+
+        norm_grad_wrt_gen_loss = grad_layer_wrt_loss(gen_loss, last_dec_layer).norm(p = 2)
+        norm_grad_wrt_perceptual_loss = grad_layer_wrt_loss(perceptual_loss, last_dec_layer).norm(p = 2)
+        adaptive_weight = safe_div(norm_grad_wrt_perceptual_loss, norm_grad_wrt_gen_loss)
+
+        # reconstruction loss
+
+        recon_loss = self.recon_loss_fn(fmap, img)
+
         # combine losses
 
-        loss = recon_loss + commit_loss + gen_loss
+        loss = recon_loss + perceptual_loss + commit_loss + adaptive_weight * gen_loss
         return loss
 
 # normalizations
@@ -489,6 +543,30 @@ class Transformer(nn.Module):
 
         return self.norm(x)
 
+# positional embedding
+
+class AxialPositionalEmbedding(nn.Module):
+    def __init__(
+        self,
+        dim,
+        *,
+        shape
+    ):
+        super().__init__()
+        self.dim = dim
+        frames, height, width = shape
+
+        self.pos_frames = nn.Parameter(torch.randn(frames, dim))
+        self.pos_height = nn.Parameter(torch.randn(height, dim))
+        self.pos_width = nn.Parameter(torch.randn(width, dim))
+
+    def forward(self):
+        pos_frames = rearrange(self.pos_frames, 'f d -> f 1 1 d')
+        pos_height = rearrange(self.pos_height, 'h d -> 1 h 1 d')
+        pos_width = rearrange(self.pos_width, 'w d -> 1 1 w d')
+        positions = pos_frames + pos_height + pos_width
+        return rearrange(positions, 'f h w d -> 1 (f h w) d')
+
 # sampling helpers
 
 def top_k(logits, thres = 0.5):
@@ -548,10 +626,7 @@ class NUWA(nn.Module):
         self.max_video_frames = max_video_frames
         video_shape = (max_video_frames, fmap_size, fmap_size)
 
-        self.video_pos_emb = AxialPositionalEmbedding(
-            dim = dim,
-            axial_shape = video_shape
-        )
+        self.video_pos_emb = AxialPositionalEmbedding(dim, shape = video_shape)
 
         self.video_transformer = Transformer(
             dim = dim,
@@ -569,6 +644,19 @@ class NUWA(nn.Module):
 
         self.to_logits = nn.Linear(dim, num_image_tokens)
 
+    def embed_text(self, text, mask = None):
+        batch, seq_len, device = *text.shape, text.device
+        assert seq_len <= self.text_max_seq_len, 'your input text has a greater length than what was designated on initialization'
+
+        tokens = self.text_embedding(text)
+        pos_emb = self.text_pos_embedding(torch.arange(seq_len, device = device))
+        tokens = tokens + rearrange(pos_emb, 'n d -> 1 n d')
+
+        return self.text_transformer(
+            tokens,
+            mask = mask
+        )
+
     @torch.no_grad()
     @eval_decorator
     def generate(
@@ -580,25 +668,18 @@ class NUWA(nn.Module):
         temperature = 1.
     ):
         batch, seq_len, device = *text.shape, text.device
-        assert seq_len <= self.text_max_seq_len, 'your input text has a greater length than what was designated on initialization'
-
-        tokens = self.text_embedding(text)
-        pos_emb = self.text_pos_embedding(torch.arange(seq_len, device = device))
-        tokens = tokens + rearrange(pos_emb, 'n d -> 1 n d')
-
-        text_embeds = self.text_transformer(
-            tokens,
-            mask = text_mask
-        )
+        text_embeds = self.embed_text(text, mask = text_mask)
 
         bos = repeat(self.video_bos, 'd -> b 1 d', b = batch)
 
         video_indices = torch.empty((batch, 0), device = device, dtype = torch.long)
         total_video_tokens = self.video_fmap_size * self.video_fmap_size * self.max_video_frames
 
-        for _ in range(total_video_tokens):
+        pos_emb = self.video_pos_emb()
+
+        for ind in range(total_video_tokens):
             frame_embeddings = self.image_embedding(video_indices)
-            frame_embeddings = self.video_pos_emb(frame_embeddings) + frame_embeddings
+            frame_embeddings = pos_emb[:, :ind] + frame_embeddings
             frame_embeddings = torch.cat((bos, frame_embeddings), dim = 1)
 
             frame_embeddings = self.video_transformer(
@@ -632,23 +713,14 @@ class NUWA(nn.Module):
         return_loss = False
     ):
         batch, seq_len, device = *text.shape, text.device
-        assert seq_len <= self.text_max_seq_len, 'your input text has a greater length than what was designated on initialization'
-
-        tokens = self.text_embedding(text)
-        pos_emb = self.text_pos_embedding(torch.arange(seq_len, device = device))
-        tokens = tokens + rearrange(pos_emb, 'n d -> 1 n d')
-
-        text_embeds = self.text_transformer(
-            tokens,
-            mask = text_mask
-        )
+        text_embeds = self.embed_text(text, mask = text_mask)
 
         frame_indices = self.vae.get_video_indices(video)
         frame_indices = rearrange(frame_indices, 'b ... -> b (...)')
         frame_indices_input = frame_indices[:, :-1] if return_loss else frame_indices
 
         frame_embeddings = self.image_embedding(frame_indices_input)
-        frame_embeddings = self.video_pos_emb(frame_embeddings) + frame_embeddings
+        frame_embeddings = self.video_pos_emb()[:, :-1] + frame_embeddings
 
         bos = repeat(self.video_bos, 'd -> b 1 d', b = batch)
         frame_embeddings = torch.cat((bos, frame_embeddings), dim = 1)
